/* fopen.c
 *
 *	(C) Copyright Apr 14 1995, Edmond J. Breen.
 *		   ALL RIGHTS RESERVED.
 * This code may be copied for personal, non-profit use only.
 * It cannot be redistributed in anyform modified or unmodified
 * without permission from the copyright holder.
 *
 *
 * 
 *
 *    ANSI C style buffed I/O routines:
 *              ungetc
 *              fopen, fputc, fflush,
 *              fclose, fgetc, ftell, fseek,
 *              fgetpos, fsetpos, fread, fwrite,
 *              fgets, fputs, ferror, feof, fprintf
 *              setvbuf, setbuf, tmpnam, freopen
 *
 * Also included are the UNIX pipe functions:
 *                popen and pclose.
 * 
 * The above functions require the UNIX style
 * low level I/O routines:
 *                      read, write, lseek,
 *                      open, close, 
 *
 *
 * Tests:
 *      these functions pass P. J Plaugers Standard C library
 *      tests:
 *                 tststdio1.c and tststdio2.c
 *      
 */

#include "stdio.h"
#include "unistd.h"
#include "fcntl.h"
#include "stdlib.h"
#include "string.h"
#include "unistd.h"

#define _READ 	 0x1	/* file open for reading */
#define _WRITE   0x2	/* file open for writing */
#define _UNBUF   0x4	/* file is unbuffered */
#define _LNBUF   0x8	/* file is line buffered  */
#define _FLBUF   0x10   /* file is fully buffered */
#define _UPDATE  0x20   /* file opened for updating */
#define _EOF     0x40	/* end of file has occured */
#define _ERR     0x80	/* error has occured */

#define _RDP    0x100  /* read has occured since last file-positioning */
#define _WRP    0x200  /* write has occured since last file-positioning */
#define _EXTBUF 0x400  /* external buffer assigned to file */

#define _UNFIL  0x800  /* file buffer not freed on close */

#define Okay2Write(fp)  !(fp->_flag &_RDP)
#define Okay2Read(fp)   !(fp->_flag &_WRP)
#define SetErr(fp)      fp->_flag |= _ERR

FILE *_LSW;                  /* last stream written too */     
int _fprep(FILE * fp);
int _fillbuff(FILE * fp);
#define PERMS 0666		/* RW for owner, group and others */
size_t _NextFopenEntry = 1;

#ifndef _EiC
FILE _IoB[FOPEN_MAX] = {
	{{NULL},           0, {NULL}, 	  _READ,           0}, /* stdin*/
        {{NULL},           0, {NULL},         _WRITE | _LNBUF, 1}, /* stdout */ 
	{{_IoB[2]._sbuf},  0, {_IoB[2]._sbuf},_WRITE | _UNBUF, 2}, /* stderr*/
};


/* methods for FILE data */

#define get_ptr(fp)    	   (fp)->_ptr.p
#define get_base(fp)   	   (fp)->_base.p
#define get_tmpnam(fp)     (fp)->_tmpnam.p
#define get_bsize(fp)  (fp)->_bsize
#define get_cnt(fp)    (fp)->_cnt
#define get_flag(fp)   (fp)->_flag
#define get_fd(fp)     (fp)->_fd
#define get_pid(fp)    (fp)->_pid   
#define get_sbuf(fp)   (fp)->_sbuf   
#define get_entry(fp)  (fp)->_entry


#endif

static unsigned char *CPTR;

void clearerr(FILE *fp)
{
    if(get_flag(fp) & (_EOF | _ERR))
	get_flag(fp) &= ~(_EOF | _ERR);
}


long ftell(FILE *fp)
{
    off_t pos = lseek(get_fd(fp),0,SEEK_CUR);
    if(pos == -1) {
	SetErr(fp);
	return EOF;
    }

    if(get_flag(fp) & _WRP)
	pos +=  ((char*)get_ptr(fp) - (char*)get_base(fp));
    else if(get_flag(fp) & _RDP)
	pos -= get_cnt(fp);

    return (long)pos;
}

char * tmpnam(char *s)
{
    /* This function came from P.J. Plauger book
     * 'The Standard C library' 1992, pg: 284.
     */
    int i;
    char *p;
    unsigned short t;
    static char buf[L_tmpnam];
    static unsigned short seed = 0;

    if(s == NULL)
	s = buf;
    seed = seed == 0 ? getpid(): seed + 1;
    strcpy(s,"/tmp/t");
    i = 5;
    p = s + strlen(s) + i;
    *p = '\0';
    for(t=seed; 0 <= --i; t>>=3)
	*--p = '0' + (t & 07);
    return s;
}
    
    

FILE *tmpfile(void)
{
    FILE * fp;
    char str[L_tmpnam];
    tmpnam(str);
    fp = fopen(str,"wb+");
    if(fp) {
	get_tmpnam(fp) = malloc(strlen(str) + 1);
	if(get_tmpnam(fp))
	    strcpy(get_tmpnam(fp),str);
	else {
	    _Ufclose(fp);
	    remove(str);
	    fp = NULL;
	}
    }
    return fp;
}
	

int setvbuf(FILE *fp, char * buf, int mode, size_t size)
{
    /* returns zero on success */
    
    if(get_base(fp))
	return -1;  /* to late, already assigned a buffer */
    switch(mode) {
      case _IOFBF: break; /* default state anyway */
      case _IOLBF: get_flag(fp) |= _LNBUF; break;
      case _IONBF: get_flag(fp) |= _UNBUF; break;
      default: return 1;
    }
    if(buf) {
	if(size) {
	    get_flag(fp) |= _EXTBUF;
	    get_base(fp) = get_ptr(fp) = buf;
	    get_bsize(fp) = size;
	} else
	    return -1;
    }
    return 0;
}
    
int fseek(FILE *fp, long offs, int orig)
{
    long ext;
    off_t lseek(int fildes, off_t offset, int whence);

    
    if(orig == SEEK_CUR && get_flag(fp) == _RDP)
	ext = offs - get_cnt(fp);
    else
	ext = offs;

    if(_Ufflush(fp) == EOF)
	return 1;  /* error */

    if(lseek(get_fd(fp),ext,orig) == EOF)
	return 1;  /* error */
    get_flag(fp) &= ~(_RDP | _WRP | _EOF);
    return 0;
}


int _fprep(FILE * fp)
{
    /*
     *  Add in the stream buffer 
     */
    if ((get_base(fp) =  malloc(BUFSIZ)) != NULL) {
	get_ptr(fp) = get_base(fp);
	get_cnt(fp) = 0;
	get_bsize(fp) = BUFSIZ;
    } else
	return 0;
    return 1;
}


int _fillbuff(FILE * fp)
{
    if ((get_flag(fp) &(_READ |_EOF | _ERR)) != _READ) 
	return EOF;
    while (get_base(fp) == NULL)
	if (get_flag(fp) & _UNBUF)
	    get_base(fp) = (char*)get_sbuf(fp);
	else if ((get_base(fp) = calloc(BUFSIZ, 1)) == NULL)
	   get_flag(fp) |= _UNBUF;
        else
	    get_bsize(fp) = BUFSIZ;
    
    get_ptr(fp) = get_base(fp);

    if(_LSW == stdout) {
	if(fp == stdin)
	    _Ufflush(stdout);
	_LSW = NULL;
    }
    
    get_cnt(fp) = read(get_fd(fp), get_ptr(fp),
		    get_flag(fp) & _UNBUF ? 1 : get_bsize(fp));
    if (get_cnt(fp) - 1 < 0) {
   	if (get_cnt(fp) - 1 == -1)
	    get_flag(fp) |= _EOF;
        else
	    SetErr(fp);

	get_cnt(fp) = 0;
	return EOF;
    }
    get_flag(fp) |= _RDP;
    return get_cnt(fp);
}


int _Uungetc(int c, FILE *fp)
{
    if(c == EOF || !(get_flag(fp) & _READ) || (get_flag(fp) & _ERR))
	return EOF;    
    if(get_cnt(fp)) {
	char *p = get_ptr(fp);
	*--p = c;
	get_ptr(fp) = p;
     } else {
	if(!get_base(fp)) {
	    if(!_fprep(fp)) {
		SetErr(fp);
		return EOF;
	    }
	} else /* redundancy check */
	    get_ptr(fp) = get_base(fp);
	
	*(char*)get_ptr(fp) = (unsigned) c;	
	get_flag(fp) &= ~_EOF;
    }
    get_cnt(fp)++;
    return c;    
}	
	
	
int fgetc(FILE * fp)
{
    
    if ((get_flag(fp) & _ERR))
	return EOF;

    if(!Okay2Read(fp) || (get_flag(fp) & _READ) == 0) {
	SetErr(fp);
	return EOF;
    }
	
    if (get_cnt(fp) == 0)
	if (_fillbuff(fp) == EOF)
	    return EOF;
    get_cnt(fp)--;
    get_flag(fp) |= _RDP;
    CPTR = get_ptr(fp);
    get_ptr(fp)= CPTR + 1;
    return *(unsigned char*)CPTR;
}

int fputc(int c, FILE * fp)
{
    if ((get_flag(fp) & _ERR))
	return EOF;

    if(!Okay2Write(fp)  || (get_flag(fp) & _WRITE) == 0) {
	SetErr(fp);
	return EOF;
    }
    
    if(get_base(fp) == NULL)
	if(!_fprep(fp)) {
	    SetErr(fp);
	    return EOF;
	}

    CPTR = get_ptr(fp);
    get_ptr(fp) = CPTR + 1;
    *CPTR = c;
    get_cnt(fp)++;
    if ((get_cnt(fp) >= get_bsize(fp)) || (get_flag(fp) & _UNBUF) ||
	(c == '\n' && get_flag(fp) & _LNBUF))
	_Ufflush(fp);
    get_flag(fp) |= _WRP;
    _LSW = fp;
    return c;
}

int fputs(const char *s, FILE *fp)
{
    if ((get_flag(fp) & _ERR))
	return EOF;

    if(!Okay2Write(fp) || (get_flag(fp) & _WRITE) == 0) {
	SetErr(fp);
	return EOF;
    }
    
    if(get_base(fp) == NULL)
	if(!_fprep(fp)) {
	    SetErr(fp);
	    return EOF;
	}

    while(*s != '\0') {
	CPTR = get_ptr(fp);
	get_ptr(fp) = CPTR + 1;
	*CPTR = *s;
	get_cnt(fp)++;
	if ((get_cnt(fp) >= get_bsize(fp)) || (get_flag(fp) & _UNBUF) ||
	      (*s == '\n' && get_flag(fp) & _LNBUF))
	    if(_Ufflush(fp) == EOF) 
		return EOF;
	++s;
    }
    get_flag(fp) |=  _WRP;
    _LSW = fp;
    return 1;
}


char * fgets(char *s, int n, FILE *fp)
{
    char  *r;
    int t;
    
    if ((get_flag(fp) & _ERR))
	return NULL;

    if(!Okay2Read(fp) || (get_flag(fp) & _READ) == 0) {
	SetErr(fp);
	return NULL;
    }
	
    
    if(get_cnt(fp) == 0)
	if(_fillbuff(fp) == EOF)
	    return NULL;    
    r = s;
    for(t = 0,--n;t<n;++t) {
	CPTR = get_ptr(fp);
	get_ptr(fp) = CPTR + 1;
	*r = *CPTR;
	get_cnt(fp)--;
	if(*r++ == '\n')
	    t = n;
	if(get_cnt(fp) == 0 && t != n)
	    if(_fillbuff(fp) == EOF)
		break;
    }
    *r = '\0';
    get_flag(fp) |= _RDP;
    return s;
}

char * gets(char *s)
{
    FILE * fp = stdin;
    char  *r;
    
    if(get_cnt(fp) == 0)
	if(_fillbuff(fp) == EOF)
	    return NULL;    
    r = s;
    while(1) {
	CPTR = get_ptr(fp);
	get_ptr(fp) = CPTR + 1;
	*r = *CPTR;
	get_cnt(fp)--;
	if(*r++ == '\n') {
	    r--;
	    break;
	}
	if(get_cnt(fp) == 0)
	    if(_fillbuff(fp) == EOF)
		break;
    }
    *r = '\0';
    get_flag(fp) |=  _RDP;
    return s;
}


size_t fread(void *buf, size_t elsize, size_t nelems, FILE * fp)
{
/*
 * (c) Edmond J. Breen (1995).
 * Reads n elements each of size elsize
 * into the buffer pointed to by buf from
 * the stream pointer fp.
 * Returns the number of items (not bytes) read.
 * or EOF on error.
 */
    size_t tsize;
    int n;
    char *s;

    if(!Okay2Read(fp)) {
	SetErr(fp);
	return EOF;
    }
    
    tsize = elsize * nelems;
    if (tsize == 0)
	return 0;
    if ((get_flag(fp) & _READ) == 0 || (get_flag(fp) & (_EOF | _ERR)) != 0)
	return 0;

    s = buf;
    while (tsize > 0) {
	if (get_cnt(fp) == 0)
	    n = _fillbuff(fp);
	else
	    n = get_cnt(fp);
	if (n == EOF)
	    break;
	if (n > tsize)
	    n = tsize;
	memcpy(s, get_ptr(fp), n);
	s += n;
	tsize -= n;
	get_cnt(fp) -= n;
	get_ptr(fp) = (char*)get_ptr(fp) + n;
    }
    get_flag(fp) |= _RDP;
    return (size_t) (s - (char *) buf) / elsize;
}

size_t fwrite(const void *buf, size_t elsize, size_t nelems, FILE * fp)
{
/*
 * (c) Edmond J. Breen  (1995).
 * Writes n elements each of size elsize
 * from the buffer buf to the stream fp.
 * Returns the number of items (not bytes) read.
 * or EOF on error.
 */
    size_t tsize, tot;
    int n;
    const char *s;


    if(!Okay2Write(fp)) {
	SetErr(fp);
	return EOF;
    }
	
    
    tsize = elsize * nelems;
    if (tsize == 0)
	return 0;
    if ((get_flag(fp) & _WRITE) == 0 || (get_flag(fp) & (_EOF | _ERR)) != 0)
	return EOF;

    s = buf;
    tot = (get_flag(fp) & _UNBUF) == 0 ? get_bsize(fp) : 1;

    if(get_base(fp) == NULL)
	if(!_fprep(fp)) {
	    SetErr(fp);
	    return EOF;
	}

    while (tsize > 0) {
	n = tot - get_cnt(fp);
	if (n > tsize)
	    n = tsize;
	memcpy(get_ptr(fp), s, n);
	s += n;
	tsize -= n;
	get_cnt(fp) += n;
	get_ptr(fp) =  (char*)get_ptr(fp) + n;
	if (get_cnt(fp) >= tot)
	    if (_Ufflush(fp) == EOF)
		return EOF;
    }
    get_flag(fp) |= _WRP;
    _LSW = fp;
    return (size_t) (s - (char *) buf) / elsize;
}



FILE *_Ufreopen(const char *name, const char *mode, FILE * fp)
{
    fclose(fp);    
    return fopen(name,mode);
}


FILE *_Ufopen(const char *name, const char *mode)
{
/*fopen function.
 *
 * (c) Edmond J. Breen (1995).
 *
 * Modes              Specification
 *  "a"                 append
 *  "r"                 reading
 *  "w"                 writing.
 *  "ab"                append binary
 *  "rb"                read binary
 *  "wb"                write binary
 *  "a+"                append update
 *  "r+"                read  update
 *  "w+"                write update
 *  "ab+" or "a+b"  append binary update
 *  "rb+" or "r+b"  read binary update
 *  "wb+" or "w+b"  write binary update
 */
    int fd, access, flag;
    FILE *fp;

    if (*mode != 'r' && *mode != 'w' && *mode != 'a')
	return NULL;
    /*
     * look for free slot.
     */
    for (fd = 0; fd < FOPEN_MAX; ++fd)
	if (_IoB[fd]._flag == 0) {
	    _IoB[fd]._entry = _NextFopenEntry++;
	    break;
	}
    if (fd >= FOPEN_MAX)
	return NULL;

    /* collect access modes and flag */
    flag = access = 0;
    while (*mode != '\0') {
	switch (*mode++) {
	case 'a':
	    flag |= _WRITE;
	    access |= (O_APPEND | O_CREAT | O_WRONLY);
	    break;
	case 'b':
	    access |= O_BINARY;
	    break;
	case '+':
	    flag |= (_WRITE | _READ | _UPDATE);
	    access = (access & ~(O_WRONLY | O_RDONLY)) | O_RDWR;
	    break;
	case 'w':
	    flag |= _WRITE;
	    access |= (O_WRONLY | O_TRUNC | O_CREAT);
	    break;
	case 'r':
	    flag |= _READ;
	    access |= O_RDONLY;
	    break;
	default:	    
	    return NULL;

	}
    }
    if ((access & O_BINARY) == 0)
	access |= O_TEXT;

    fp = &_IoB[fd];

    fd = open(name, access, PERMS);
    if (fd == -1)
	return NULL;
    get_fd(fp) = fd;
    get_flag(fp) = flag;
    get_ptr(fp) = get_base(fp) = NULL;
    get_tmpnam(fp) = NULL;
    return fp;
}


FILE *popen(const char *name, const char *mode)
{
/* popen function.
 *
 * (c) Edmond J. Breen (1995).
 *    All rights reserved.
 *
 * Modes              Specification
 *  "r"                 reading
 *  "w"                 writing.
 */

#define _SHELL  "/bin/sh"
#define tst(a,b)  (access == O_RDONLY ? (b) : (a))
    
    int fd, flag, access, p[2];
    FILE *fp;

    if (*mode != 'r' && *mode != 'w')
	return NULL;
    else { /* collect access and flag mode */
	if(*mode == 'r') {
	    flag = _READ;
	    access = O_RDONLY;
	} else {
	    flag = _WRITE;
	    access = O_WRONLY;
	}
    }
    /* open pipe */
    if(pipe(p) < 0)
	return NULL;
    /*
     * look for free slot.
     */
    for (fd = 0; fd < FOPEN_MAX; ++fd)
	if (_IoB[fd]._flag == 0) {
	    _IoB[fd]._entry = _NextFopenEntry++;
	    break;
	}
    if (fd >= FOPEN_MAX)
	return NULL;

    fp = &_IoB[fd];

    if((get_pid(fp) = fork()) == 0) {  /* in child */
	close(tst(p[O_WRONLY],p[O_RDONLY]));
	close(tst(O_RDONLY,O_WRONLY));
	dup(tst(p[O_RDONLY],p[O_WRONLY]));
	close(tst(p[O_RDONLY],p[O_WRONLY]));
	
        /* close all descriptors in _IoB: POSIX.2 compliant */
	for (fd = 0; fd < FOPEN_MAX; ++fd)
	    if (_IoB[fd]._flag != 0 && _IoB[fd]._pid > 0)
		close(_IoB[fd]._fd);
	
	execl(_SHELL,"sh","-c",name,0);
	return NULL;
    }
    
    if(get_pid(fp) == -1)
	return NULL;
    
    close(tst(p[O_RDONLY],p[O_WRONLY]));

    get_fd(fp) = tst(p[O_WRONLY],p[O_RDONLY]);
    get_flag(fp) = flag;
    return fp;
#undef tst
#undef _SHELL
}

void _ffexit(size_t Entry)
{
    int i;
    for (i = 0; i < FOPEN_MAX; i++)
	if ((_IoB[i]._flag &  _IoB[i]._entry) >= Entry)
	    _Ufclose(&_IoB[i]);
}

int _Ufflush(FILE * fp)
{
    if (fp == NULL) {		/* flush all output streams */
	int i;
	for (i = 0; i < FOPEN_MAX; i++)
	    if (_IoB[i]._flag & _WRITE)
		_Ufflush(&_IoB[i]);
	return 0;
    }

    if (get_cnt(fp) == 0 || get_base(fp) == NULL || (get_flag(fp) & _ERR) != 0 )
	return 0;
    if( get_flag(fp) & _WRITE && !(get_flag(fp) & _RDP)) {
	get_cnt(fp) = write(get_fd(fp), get_base(fp),
			    (char*)get_ptr(fp) - (char*)get_base(fp));
	if (get_cnt(fp) <= 0) {	/* report error */
	    SetErr(fp);
	    return EOF;
	}
    }
    get_ptr(fp) = get_base(fp);
    get_cnt(fp) = 0;
    get_flag(fp) &=  ~(_WRP | _RDP);
    return 0;
}

int _Ufclose(FILE * fp)
{
    _Ufflush(fp);
    if ((get_flag(fp) & (_UNBUF | _EXTBUF)) == 0) {
	if(get_base(fp))
	    free(get_base(fp));
	get_ptr(fp) = get_base(fp) = NULL;
    }

    if(close(get_fd(fp)) < 0)
	return EOF;
    if(get_tmpnam(fp)) {
	remove(get_tmpnam(fp));
	free(get_tmpnam(fp));
	get_tmpnam(fp) = NULL;
    }
    get_cnt(fp) = get_flag(fp) = 0;
    return 0;
}

#include <errno.h>
int pclose(FILE *fp)
{
    int status;
    pid_t waitpid(pid_t pid, int *status, int options);

    if(_Ufclose(fp) == EOF)
	return -1;

    while(waitpid(get_pid(fp), &status,0) < 0)
	if(errno != EINTR)
	    return -1;

    return status;
}

















