/* func.c
 *
 *	(C) Copyright Nov 23 1996, Edmond J. Breen.
 *		   ALL RIGHTS RESERVED.
 * This code may be copied for personal, non-profit use only.
 *
 *
 * This file contains the methods for dealing with EiC function
 * house keeping.
 */

#include "stdlib.h"
#include "stdio.h"
#include "string.h"

#include "global.h"
#include "typemod.h"
#include "func.h"
#include "xalloc.h"
#include "symbol.h"

int gotMissingNames(func_t *f)
{
    /* check to see that all parameters have a name */
    int i;

    if(!hasPrototype(f) || gettype(getFPty(f, 0)) == t_void)
	return 0;

    for (i = 0; i < getFNp(f); ++i)
	if(!getFPname(f,i)) 
	    if(gettype(getFPty(f, i)) != t_var)
	       return i+1;
    return 0;
}

int IsVariadic(func_t *f)
{
    return f ? getFNp(f) > 1 && gettype(getFPty(f,getFNp(f)-1)) ==
	t_var : 0;
}

int hasPrototype(func_t *f)
{
    
    return f ? gettype(getFPty(f, 0)) != t_var: 0;
}

void swapFPLists(func_t *f1, func_t *f2)
{
    /* swap parameter lists */
    type_expr **t = getFPtyList(f1);
    char  **n = getFPnames(f1);
    char  **c = getFPcomms(f1);
    
    setFPtyList(f1,getFPtyList(f2));
    setFPnames(f1,getFPnames(f2));
    setFPcomms(f1,getFPcomms(f2));
    
    setFPtyList(f2,t);
    setFPnames(f2,n);
    setFPcomms(f2,c);
}


int HasHiddenParm(type_expr *ty)
{
    int t = gettype(nextType(ty));
    if(t == t_struct || t == t_union)
	return 1;
    return 0;
}

int sameFuncs(type_expr * t1, type_expr * t2)
{
    /* check for compatible function types */

    func_t * f1, *f2;
    int i;

    f1 = getInf(t1);
    f2 = getInf(t2);
    if(!f1 || !f2)   /* watch for builtin types before prototype */
        return 0;
    if(hasPrototype(f1) && hasPrototype(f2)) {
        if (getFNp(f1) != getFNp(f2))
            return 0;
        for (i = 0; i < getFNp(f1); ++i)
            if(!sametypes(getFPty(f1,i),getFPty(f2,i)))
                return 0;
    } else if(IsVariadic(f1) || IsVariadic(f2))
        return 0;
    else { /* now check parameters are not convertable */
        if(hasPrototype(f2))
            f1 = f2;
        for (i = 0; i < getFNp(f1); ++i)
            switch(gettype(getFPty(f1,i))) {
              case t_char: case t_uchar:
              case t_short: case t_ushort:
              case t_float:
                return 0;
            }
    }
    /* check function return types */
    return sametypes(t1->nxt, t2->nxt);
}


void make_func(token_t * e1)
{
    /*    if (gettype(e1->type) == t_array ||
	gettype(e1->type) == t_funcdec) {
	    error("Illegal type construction");
	    }
	    */
    e1->type = addtype(t_funcdec, e1->type);
    setInf(e1->type,xcalloc(1, sizeof(func_t)));
}

void adjustParam(type_expr **type)
{
    int t = gettype(*type);
    if(t == t_funcdec)
	*type = addtype(t_pointer, *type);
    else if(t == t_array)
	exchtype(t_pointer,*type);
}

void newParam(func_t *f,char *name, type_expr *type)
{
    setFPtyList(f,xrealloc(getFPtyList(f),
			  sizeof(type_expr *) * (getFNp(f) +1)));
    setFPnames(f,xrealloc(getFPnames(f),
			  sizeof(char *) * (getFNp(f) +1)));
    if(name)
	setFPname(f,getFNp(f), strsave(name));
    else
	setFPname(f,getFNp(f), NULL);
    setFPty(f,getFNp(f),copytype(type));
}

void add_func_parm(func_t * f, type_expr ** type, char * name)
{
    int t;
    type_expr *t1, *t2;
    /* must handle storage class, as yet not done. */
    if ((t = gettype(*type)) != t_funcdec && t != t_array)
	t = 0;

    newParam(f,name,*type);

    /*
     * Because, parameters will be removed
     * from the lookup table the function types stored
     * need to be as complete as possible.
     */
    t1 = getFPty(f,getFNp(f));
    t2 = *type;
    while (t2) {
	if (!t2->alias) {
	    t2->alias = 1;
	    t1->alias = 0;
	}
	t2 = nextType(t2);
	t1 = nextType(t1);
    }

    adjustParam(&getFPty(f,getFNp(f)));
    setFNp(f,getFNp(f)+1);
}

void add_func_str(func_t * f, char *s)
{
    if (!f->sn)
	f->strs = (char **) xcalloc(1, sizeof(char *));
    else
	f->strs = (char **) xrealloc(f->strs,
				     sizeof(char *) * (f->sn + 1));
    f->strs[f->sn] = s;
    f->sn++;
}


void add_func_static(func_t * f, int n)
{
    if (!f->stn)
	f->st = (int *) xcalloc(1, sizeof(int));
    else
	f->st = (int *) xrealloc(f->st,
				     sizeof(int) * (f->stn + 1));
    f->st[f->stn] = n;
    f->stn++;
}

void add_func_initialiser(func_t * f, void *s)
{
    if (!f->ni)
	f->inzs = (void **) xcalloc(1, sizeof(void *));
    else
	f->inzs = (void **) xrealloc(f->inzs,
				     sizeof(void *) * (f->ni + 1));
    f->inzs[f->ni] = s;
    f->ni++;
}

void freeFuncComments(func_t *f)
{
    int i;
    if(f->pcomms) {
	for(i=0;i<getFNp(f);++i)
	    if(getFPcomm(f,i))
		xfree(getFPcomm(f,i));
	xfree(getFPcomms(f));
	setFPcomms(f,NULL);
    }
    if(getFComm(f)) {
	xfree(getFComm(f));
	setFComm(f,NULL);
    }
}		    
		    

void free_func_inf(func_t * f)
{
    int i;
    if (!f)
	return;
    if (getFNp(f)) {
	for (i = 0; i < getFNp(f); i++) {
	    freetype(getFPty(f,i));
	    if(getFPname(f,i))
		xfree(getFPname(f,i));
       }
	    
	xfree(getFPtyList(f));
	xfree(getFPnames(f));
    }
    if (f->sn) {
	for (i = 0; i < f->sn; ++i)
	    xfree(f->strs[i]);
	xfree(f->strs);
	f->sn = 0;
    }
    if (f->ni) {
	for (i = 0; i < f->ni; ++i)
	    xfree(f->inzs[i]);
	xfree(f->inzs);
	f->ni= 0;
    }
    if (f->stn) {
	for (i = 0; i < f->stn; ++i)
	    xfree(ENV->AR[f->st[i]].v.p.p);
	xfree(f->st);
	f->stn = 0;
    }
    freeFuncComments(f);
    xfree(f);
}

void markFunc(type_expr * type, int mark)
{
    func_t * f = getInf(type);
    int i;
    
    xmark(f,mark);
    /* mark paramaters */
    if(getFNp(f)) {
	xmark(getFPtyList(f),mark);
	xmark(getFPnames(f),mark);
	if(getFPcomms(f))
	    xmark(getFPcomms(f),mark);
	for(i=0;i<getFNp(f);i++) {
	    marktype(getFPty(f,i),mark);
	    if(getFPname(f,i))
		xmark(getFPname(f,i),mark);
	    if(f->pcomms && f->pcomms[i])
		xmark(f->pcomms[i],mark);
	}
    }
    /* mark strings */
    if(f->sn) {
	xmark(f->strs,mark);
	for(i=0;i<f->sn;i++) 
	    xmark(f->strs[i],mark);
		
    }
    /* mark initializers */
    if(f->ni) {
	xmark(f->inzs,mark);
	for(i=0;i<f->ni;i++)
	    xmark(f->inzs[i],mark);
    }
     /* mark static arrays */
    if(f->stn) {
	xmark(f->st,mark);
	for(i=0;i<f->stn;i++)
	    xmark(ENV->AR[f->st[i]].v.p.p,mark);
    }
    if(getFComm(f))
	xmark(getFComm(f),mark);
}

void add_func_comment(char *s)
{
    extern token_t *RETURNEXPR;
    func_t * f = getInf(RETURNEXPR->type);
    setFComm(f,s);
}

void (*saveComment()) (char * s)
{
    extern int INFUNC;
    extern int INPARMLIST;
    if(INFUNC) {
	 extern token_t *RETURNEXPR;
	 func_t * f = getInf(RETURNEXPR->type);
	 if(f && !getFComm(f))
	     return add_func_comment;
    } 	
    return NULL;
}
