/* lexer.c
 *
 *	(C) Copyright Apr 15 1995, Edmond J. Breen.
 *		   ALL RIGHTS RESERVED.
 * This code may be copied for personal, non-profit use only.
 *
 *
 * History:
 *           First written V1.0 1984 (E.J.B)
 *           Revised       V1.2 1994 (E.J.B)
 *           Revised       V2.0 1995 (E.J.B)
 *
 */
 
#include "stdio.h"
#include "stdlib.h"
#include "limits.h"
#include "string.h"
#include "math.h"
#include "float.h"

#include "lexer.h"
#include "global.h"
#include "xalloc.h"
#include "error.h"
#include "symbol.h"

char nextchar(void);
char *nextproline(void);
void retract(char c);
int fail(int ival, int c);
int charliteral(int c);
void stringliteral(void);
void success(int ival);


#define NEWLINE(ch)  (ch  == '\n' ? TRUE: FALSE)
#define LETTER(ch)   ((ch >=  'a'  && ch   <=  'z') ? TRUE :\
				 ((ch >=  'A'  && ch   <=  'Z') ? TRUE :\
				 ((ch ==  '_'   || ch   ==  '$')? TRUE : FALSE)))

#define DIGIT(ch)   ((ch  >=  '0'   && ch   <= '9') ? TRUE : FALSE)
#define WHITE(ch)   ((ch  == ' '    || ch   == '\t')? TRUE : FALSE)



keyword_t cwords[] =
{
  {"auto", autosym,},
  {"break", breaksym,},
  {"case", casesym,},
  {"char", charsym,},
  {"const", constsym,},
  {"continue", continuesym,},
  {"default", defaultsym,},
  {"do", dosym,},
  {"double", doublesym,},
  {"else", elsesym,},
  {"enum", enumsym,},
  {"extern", externsym,},
  {"float", floatsym,},
  {"for", forsym,},
  {"if", ifsym,},
  {"int", intsym,},
  {"long", longsym,},
  {"register", registersym,},
  {"return", returnsym,},
  {"short", shortsym,},
  {"signed", signedsym,},
  {"sizeof", sizeofsym,},
  {"static", staticsym,},
  {"struct", structsym,},
  {"switch", switchsym,},
  {"typedef", typedefsym,},
  {"union", unionsym,},
  {"unsigned", unsignedsym,},
  {"void", voidsym,},
  {"volatile", volatilesym,},
  {"while", whilesym,},
};



#define NOTOKEN 0

char LEXEM[BSIZE+1];
token_t token;

#ifdef ILOOKAHEAD

token_t TokenArray[5];
int TokenI = 0;

#else

unsigned short STOKEN;

#endif




unsigned lex_lineno, lex_lineprev, lex_linepos, lex_linelen, lex_curpos = 0,
 lex_lastpos;
char *lex_buff;

static state = 0;
static unsigned long lexival;
static double lexfval;
static char Lseen,  /* long seen */
            Fseen,  /* floating point seen */
            Useen,  /* Unsigned specifier seen */
            Hseen;  /* Hex or Octal value seen */

#define getoct(x)  (((x)>='0'&&(x)<='7')? (x)-'0':-1)


int gethex(int c)
{
    if (c >= '0' && c <= '9')
	return c - '0';
    if (c >= 'a' && c <= 'f')
	return c - 'a' + 10;
    if (c >= 'A' && c <= 'F')
	return c - 'A' + 10;
    return -1;
}


void initlex(char *str)
{
    lex_buff = str;
    lex_curpos = 0;
    lex_lineprev = 0;
    lex_lineno = 1;

#ifdef ILOOKAHEAD
    TokenI = 0;    
#else

    STOKEN = NOTOKEN;
#endif
    
}

char nextchar()
{
    char *nextproline();
    if (lex_buff[lex_curpos] == EOF)
	return 0;
    else if (lex_buff[lex_curpos] == '\0') {
	if (lex_curpos != lex_lastpos)
	    return 0;
	lex_buff = nextproline();
	lex_lastpos = lex_curpos = 0;
	lex_lineprev = 0;
	lex_lineno = 1;
    }
    if (lex_lineprev != lex_lineno) {
	lex_linepos = lex_curpos;
	lex_lineprev = lex_lineno;
    }
    return (lex_buff[lex_curpos++]);
}

void checkExt(int c)
{				/* check for unsigned and long suffix */
    Lseen = Useen = Fseen = 0;
    if (c == 'f' || c == 'F')
	Fseen = 1;
    else if (c == 'u' || c == 'U') {
	Useen = 1;
	if ((c = nextchar()) == 'l' || c == 'L')
	    Lseen = 1;
	else
	    retract(c);
    } else if (c == 'l' || c == 'L')
	Lseen = 1;
    else
	retract(c);
}

void retract(char c)
{
    if (c != '\0') {
	lex_curpos--;
	if (lex_curpos < lex_lastpos)
	    lex_lastpos = lex_curpos;
    }
}


int fail(int ival, int c)
{
    retract(c);
    switch (ival) {
      case RELOP: return (10);
      case ID:    return (20);
      case FLOAT:
      case INT: return (100);
    }
    return 0;
}
void setfloatval(void)
{
    /*
    if (Lseen || lexfval > FLT_MAX  || !Fseen ) {
	token.val.dval = lexfval;
	token.tok = DOUBLE;
    } else {
	token.val.fval = lexfval;
	token.tok = FLOAT;
	}
	*/
    
    token.val.dval = lexfval;
    token.tok = FLOAT;
}

void setintval(void)
{
    if (Fseen) {
	lexfval = lexival;
	setfloatval();
	return;
    }
    if (Useen) {
	if (Lseen || lexival > UINT_MAX) {
	    token.tok = ULONG;
	    token.val.ulval = lexival;
	} else {
	    token.tok = UINT;
	    token.val.uival = (unsigned) lexival;
	}
    } else if (Lseen || lexival > UINT_MAX) {
	if (lexival > ULONG_MAX) {
	    token.tok = ULONG;
	    token.val.ulval = lexival;
	} else {
	    token.val.lval = lexival;
	    token.tok = LONG;
	}
    } else {
	if (lexival <= INT_MAX) {
	    token.val.ival = (int) lexival;
	    token.tok = INT;
	} else if(Hseen && lexival <= UINT_MAX) {
	    token.tok = UINT;
	    token.val.uival = (unsigned) lexival;
	} else if(lexival <= LONG_MAX) {	    
	    token.val.lval = (long) lexival;
	    token.tok = LONG;
	} else {
	   token.tok = ULONG;
	   token.val.ulval = lexival;
       }
    }
}

void success(int ival)
{
  int i, size;

  size = (int) (lex_curpos - lex_lastpos);
  memcpy(LEXEM, &lex_buff[lex_lastpos], size);
  LEXEM[size] = '\0';
  if (Lseen) size--;
  if (Useen) size--;
  if (Fseen) size--;

  Hseen = 0;
  switch (ival) {
  case ID:
    if ((token.tok = iskeyword(cwords, LEXEM,
			     sizeof(cwords) / sizeof(keyword_t))) == 0) {
      token.tok = ID;
      /* search for id in various name spaces */
      if ((token.val.sym = lookup(work_tab, LEXEM)) == NULL)
	token.val.sym = insertLUT(work_tab, LEXEM, ID);
      if (token.val.sym)
	if (token.val.sym->sclass == c_typedef)
	token.tok = TYPENAME;
  }
    break;
  case OCTAL:
    if (Fseen)
      error("Declaration syntax error");
    for (lexival = 0, i = 0; i < size; i++)
	lexival = lexival * 8 + getoct(LEXEM[i]);
    Hseen = 1;
    setintval();
    break;
  case HEX:
    for (lexival = 0, i = 2; i < size; i++)
	lexival = lexival * 16 + gethex(LEXEM[i]);
    Hseen = 1;
    setintval();
    break;
  case INT:
    for (lexival = 0, i = 0; i < size; i++)
      lexival = lexival * 10 + LEXEM[i] - '0';
    setintval();
    break;
  case FLOAT:
    if (Useen)
      error("Declaration syntax error");
    lexfval = atof(LEXEM);
    setfloatval();
    break;
  case RELOP:
  case MISC:
    break;
  }
}


int WASLITERAL;
int charliteral(int c)
{
    if (c == '\\') {
	switch ((c = nextchar())) {
	  case 'n': c = '\n'; break;     /* newline */
	  case 't': c = '\t'; break;     /* tabspace */
	  case 'v': c = '\v'; break;     /* vertical tab */
	  case 'b': c = '\b'; break;     /* backspace */
	  case 'r': c = '\r'; break;     /* carriage return */
	  case 'f': c = '\f'; break;     /* formfeed */
	  case 'a': c = '\a'; break;     /* bell */
	  case '\\': c = '\\'; break;    /* backslash */
	  case '\'': c = '\''; break;    /* single quote */
	  case '"': c = '\"'; break;     /* double quote */
	  case 'x':			   /* string of hex characters */
	  case 'X':{
	      int i, val = 0;
	      while ((i = gethex((c = nextchar()))) > -1) {
		  val = val * 16 + i;
	      }
	      retract(c);
	      if (val > 255)
		  error("Illegal character hex value");
	      c = val;
	  }
	    break;
	  default:
	    if (getoct(c) > -1) {		/* octal characters */
		int i, val = 0;
		while ((i = getoct(c)) > -1) {
		    val = val * 8 + i;
		    c = nextchar();
		}
		retract(c);
		if (val > 255)
		    error("Illegal character octal value");
		c = val;
	    } else
		error("Illegal character escape sequence `\\%c'", c);
	    break;
	}
	WASLITERAL = 1;
    } else
	WASLITERAL = 0;
    return ((signed char )c);
}

void stringliteral(void)
{
    unsigned size, lastsize = 0, c;
    char *p;
    lex_lastpos = lex_curpos;
    do {

	for (size = 0; ((c = charliteral(nextchar())) != '\0' || WASLITERAL) &&
	     !(c == '"' && !WASLITERAL)  && size < BSIZE; size++) 
	    LEXEM[size] = c;
    

	if (lastsize)
	    p = (char *) xrealloc(p, lastsize + size + 1);
	else
	    p = (char *) xcalloc(size + 1, sizeof(char));

	memcpy(&p[lastsize], LEXEM, size);
	lastsize += size;

	if(c != '"' && size == BSIZE) {
	    p[lastsize++] = c;
	    continue;
	}
	
	if (c != '"')
	    error("String literal error");

	do {
	    c = nextchar();
	    if (c == '\n')
		lex_lastpos++, lex_lineno++;
	} while (WHITE(c) || c == '\n');

	lex_lastpos = lex_curpos;
	if (!c)
	    do
		c = nextchar();
	    while (WHITE(c));
    } while (c == '"' || size == BSIZE);
    retract(c);
    p[lastsize] = '\0';
    token.val.p.sp = token.val.p.p = p;
    token.val.p.ep = p + lastsize;
}

int lexan(void)
{
    int t; char c, nextchar();

#ifdef ILOOKAHEAD

    if(TokenI) {
      token = TokenArray[--TokenI];
      return token.tok;
    }

#else

    if (STOKEN != NOTOKEN) {
	STOKEN = NOTOKEN;
	return token.tok;
    }

#endif
    
    state = 0;
    while (1) {
	switch (state) {
	  case 0: lex_lastpos = lex_curpos; c = nextchar();
	    state = (WHITE(c) ? 0 :
		    (c == '\n' ? lex_lineno++, 0 :
		    (c == '<' ? t = LT, 1 :
		    (c == '>' ? t = GT, 2 :
		    (c == '+' ? t = '+', 3 :
		    (c == '-' ? t = '-', 4 :
		    (c == '|' ? t = BOR, 5 :
		    (c == '&' ? t = AND, 6 :
		    (c == '\''? 7 :
		    (c == '"' ? 8 :
		    (c == '.' ? 9 :  
		    (c == '/' ? t = '/', c = nextchar(), 50 :
		    (c == '%' ? t = '%', c = nextchar(), 50 :
		    (c == '*' ? t = '*', c = nextchar(), 50 :
		    (c == '=' ? t = ASS, c = nextchar(), 50 :
		    (c == '!' ? t = NOT, c = nextchar(), 50 :
		    (c == '^' ? t = XOR, c = nextchar(), 50 :
		     fail(RELOP, c))))))))))))))))));
	    break;
	  case 1: /* get <,  <= and << */
	    if ((c = nextchar()) == '<') t = LSHT;
	    else state = 50;
	    break;
	  case 2: /* get >, >= and >> */
	    if ((c = nextchar()) == '>') t = RSHT;
	    else state = 50;
	    break;
	  case 3: c = nextchar();                         /* get +, += or ++ */
	    if (c == '+') t = INC, state = 60;
	    else state = 50;
	    break;
	  case 4: c = nextchar();                            /* get -, -= -- */
	    state = 60;
	    if (c == '-') t = DEC;
	    else if (c == '>') t = RARROW;
	    else state = 50;
	    break;
	  case 5: c = nextchar();                         /* get |, |= or || */
	    if (c == '|') t = LOR, state = 60;
	    else state = 50;
	    break;
	  case 6: c = nextchar();                         /* get &, &= or && */
	    if (c == '&') t = LAND, state = 60;
	    else state = 50;
	    break;
	  case 7:token.val.ival = charliteral(nextchar()); /* char_constants */
	    t = CHAR;
	    if (nextchar() != '\'')
		error("Missing single quote '");
	    state = 60;
	    break;
	  case 8: stringliteral();                        /* string literals */
	    token.tok = STR;
	    return STR;
	  case 9: c = nextchar();
	    t = '.';
	    if(DIGIT(c)) 
		state = 22;
	    else
		state = 60;
	    retract(c);
	    break;
	  case 10: c = nextchar();              /* identifiers and  keywords */
	    state = (LETTER(c) ? 11 :
		    (c == '_' ? 11 : fail(ID, c)));
	    break;
	  case 11: c = nextchar();
	    state = (LETTER(c) ? 11 :
		    (DIGIT(c) ? 11 :
		    (c == '_' ? 11 : 12)));
	    break;
	  case 12: retract(c); success(ID); return (token.tok);

	  case 20: c = nextchar();                     /* integers and reals */
	    state = (c == '0' ? 30 :
		    (DIGIT(c) ? 21 : fail(INT, c)));
	    break;
	  case 21: c = nextchar();
	    state = (DIGIT(c) ? 21 :
		    (c == '.' ? 22 :
		    (c == 'e' ? 23 :
		    (c == 'E' ? 23 : 25))));
	    break;
	  case 22: c = nextchar();
	    state = (DIGIT(c) ? 22 :
		    (c == 'e' ? 23 :
		    (c == 'E' ? 23 : 26)));
	    break;
	  case 23: c = nextchar();
	    state = (c == '+' ? 24 :
		    (c == '-' ? 24 :
		    (DIGIT(c) ? 24 : fail(FLOAT, c) /* ??? */ )));
	    break;
	  case 24: c = nextchar();
	    state = (DIGIT(c) ? 24 : 26);
	    break;
	  case 25: checkExt(c); success(INT); return (token.tok);
	  case 26: checkExt(c); success(FLOAT); return (token.tok);
	  case 27: checkExt(c); success(HEX);   return (token.tok);
	  case 28: checkExt(c); success(OCTAL); return (token.tok);
	  case 30:			  /* check for octal and hex numbers */
	    if ((c = nextchar()) == 'x' || c == 'X') {
		while (gethex((c = nextchar())) > -1);
		state = 27;
		break;
	    }
	    if (c != '.' && c != 'e' && c != 'E') {
		while (getoct(c) > -1)
		    c = nextchar();
		state = 28;
		break;
	    }
	    retract(c); state = 21; break;
	  case 50:                                      /* mix with equal's  */
	    if (c == '=')
		switch (t) {
		  case '+': t = ADDEQ;  break;		/* += */
		  case '-': t = SUBEQ;  break;		/* -= */
		  case '/': t = DIVEQ;  break;		/* /= */
		  case '*': t = MULEQ;  break;		/* *= */
		  case '%': t = MODEQ;  break;		/* %= */
		  case ASS: t = EQ;     break;		/* == */
		  case GT:  t = GE;     break;		/* >= */
		  case LT:  t = LE;     break;		/* <= */
		  case NOT: t = NE;     break;		/* != */
		  case RSHT:t = RSHTEQ; break;		/* >>= */
		  case LSHT:t = LSHTEQ; break;		/* <<= */
		  case AND: t = ANDEQ;  break;		/* &= */
		  case BOR: t = BOREQ;  break;		/* |= */
		  case XOR: t = XOREQ;  break;		/* ^= */
		  default: retract(c);
	    } else retract(c);
	    state = 60;
	    break;
	  case 60: success(MISC); token.tok = t; return (token.tok);
	  case 100: token.tok = nextchar(); return (token.tok);
	}
    }
    return token.tok;
}
